using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Amaimon {


    ///////////////////////////////////////////////////
    // Interpolable interface 
    // for things that need to be updated per frame
    ///////////////////////////////////////////////////

    public abstract class LifecycleInterpolable {
        // 0 - off/stopped , 1 - turning on, 2 - on, 3 - turning off
        protected byte _phase = 0;
        public bool waitingToBeRun = true;

        public virtual void runNow() {
            waitingToBeRun = false;
            _phase = 1;
        }
        public abstract void onTick(float deltaTimeS);
        public virtual bool getIsTicking() {
            return !waitingToBeRun && _phase > 0;
        }
        public abstract void stop();
        public virtual bool stoppedOrStopping() {
            return (_phase == 0 || _phase == 3);
        }

    }

    ///////////////////////////////////////////////////
    // InterpolablePack 
    // a bundle of interpolables that happen together
    ///////////////////////////////////////////////////

    public class InterpolablePackGen<T> : LifecycleInterpolable where T : LifecycleInterpolable {
        public List<T> list = new List<T>();

        override public void runNow() {
            waitingToBeRun = false;
            _phase = 1;
            list.ForEach(e => e.runNow());
        }
        override public void onTick(float deltaTimeS) {
            list.Where(i => i.getIsTicking()).ForEach(i => i.onTick(deltaTimeS));
        }
        public bool getIsTicking() {
            Log.d("is ticking w ip");
            return list.Any(element => element.getIsTicking());
        }
        override public void stop() {
            list.ForEach(e => e.stop());
        }

        public bool stoppedOrStopping() {
            return list.All(element => element.stoppedOrStopping());
        }

        public void purge() {
            list = list.FindAll(e => e.getIsTicking() || e.waitingToBeRun);
        }

        public void clear() {
            list.Clear();
        }
        public InterpolablePackGen<T> add(T i) {
            purge();
            list.Add(i);
            return this;
        }
    }

    public class InterpolablePack : InterpolablePackGen<LifecycleInterpolable> {
        public static InterpolablePack create() {
            return new InterpolablePack();
        }
    };

    public class ConcurrentInterpolablePack : InterpolablePackGen<OnOffInterpolator> {
        public static ConcurrentInterpolablePack create() {
            return new ConcurrentInterpolablePack();
        }

        public void setState(bool st) {
            list.ForEach(it => it.setState(st));
        }
    }

    ///////////////////////////////////////////////////
    // OscillatingInterpolable 
    // An interpolable that has a period
    ///////////////////////////////////////////////////

    public abstract class OscillatingInterpolable : LifecycleInterpolable {
        protected float _elapsed = 0; // current clock
        protected float _sustain;
        protected float _period;

        private float _newPeriod = 0f;
        public float periodsSoFar = 0f;

        public float phaseTrigger = 0.5f;
        public bool phaseTriggered = false;
        float prevPhase = 0f;

        public OscillatingInterpolable(float perdiodSecs, float sustainCycles = 100000) {
            _period = perdiodSecs;
            _sustain = sustainCycles;
        }

        public float getDuration() {
            return _period * _sustain;
        }

        override public void onTick(float deltaTimeS) {
            if(!getIsTicking())
                return;

            _elapsed += deltaTimeS;

            periodsSoFar = _elapsed/_period;
            var phase = (360 * periodsSoFar) % 360;

            var shiftedPhase = (phase>180) ? phase-360 : phase;

            if((prevPhase<phaseTrigger && phaseTrigger<=shiftedPhase)) {
                phaseTriggered = true;
            }

            prevPhase = shiftedPhase;

            if(phase < 5 && _newPeriod !=0f) {
                _elapsed = deltaTimeS;
                _period = _newPeriod;
                _newPeriod = 0f;
                periodsSoFar = _elapsed/_period;
            }

            if(periodsSoFar >= _sustain && _newPeriod == 0) {
                _phase = 0;
                onPhase(0);
            }
            else {
                onPhase(phase);
            }            
        }

        override public void stop() {
            _phase = 0; // no stopping phase
        }


        abstract public void onPhase(float phase);

        public void changePeriod(float newPeriod) {
            if(newPeriod != _period)
                _newPeriod = newPeriod;
        }

    }


    ///////////////////////////////////////////////////
    // NoisyInterpolator 
    // interpolates a value with noise function
    ///////////////////////////////////////////////////

    public abstract class NoisyInterpolator : LifecycleInterpolable {
        float _amount;
        float _speed;
        protected float _default;

        static int index = 0;

        public NoisyInterpolator(float amount, float speed) {
            _amount = amount;
            _speed = speed;
        }

        override public void onTick(float deltaTimeS) {
            if(!getIsTicking())
                return;

            float rn = Mathf.PerlinNoise(Time.time * _speed, index);

            apply(_default + rn * _amount);

            index += 100;
        }
        override public void stop() {
            _phase = 0; // no stopping phase
        }

        ~NoisyInterpolator() {
            apply(_default);
            index = 0;
        }


        abstract public void apply(float v);
    }

    ///////////////////////////////////////////////////
    // SinusaidalInterpolator
    // interpolates a value with sinus function
    ///////////////////////////////////////////////////

    abstract public class SinusaidalInterpolator : OscillatingInterpolable {
        
        public float _amplitude;
        public SinusaidalInterpolator(float amplitude, float perdiodSecs, float sustainCycles = 100000) : base(perdiodSecs, sustainCycles) {
            _amplitude = amplitude;
        }

        override public void onPhase(float phase) 
        {
            var rad =  phase * Mathf.Deg2Rad;
            var value = (float)(_amplitude * System.Math.Sin(rad));
            apply(value);
        }

        abstract public void apply(float v);


    }

    ///////////////////////////////////////////////////
    // SmoothEnvelopeInterpolator 
    // interpolates a value up to target over time when enabled
    // and back to default when disabled
    // might be disabled automatically on request
    ///////////////////////////////////////////////////

    public class LambdaEnvelopeInterpolator : SmoothEnvelopeInterpolator {
        Action<float> lambda;
        public LambdaEnvelopeInterpolator(float current, float target, float inTime, float outTime, float sustain, Action<float> l) : base (target, inTime, outTime, sustain) {
            _default = current;
            lambda = l;
        }

        override public void apply(float v) {
            lambda(v);
        }

    }

    public abstract class SmoothEnvelopeInterpolator : LifecycleInterpolable {
        protected float _default; // the value to restore when this interpolator not in use
        protected float _inTime; // enable time
        protected float _outTime; // disable time
        protected float _target; // target value
        protected float _elapsed = 0; // current clock
        protected float _current = 0; // current value
        protected float _reached = 0; // reached value if in phase was aborted
        protected float _sustain;

        public SmoothEnvelopeInterpolator(float target, float inTime, float outTime, float sustain = 0f) {
            _target = target;
            _inTime = inTime;
            _outTime = outTime;
            _sustain = sustain;
            _elapsed = 0;
        }

        ~SmoothEnvelopeInterpolator() {
            apply(_default);
        }

        public float duration() {
            return _inTime + _outTime + _sustain;
        }

        
        override public void stop() {
            _reached = _current;
            _phase = 3; // begin fade out
            _elapsed = 0f;
        }
        public void fitInto(float fullDuration) {
            _sustain = fullDuration - _outTime - _inTime;
            if(_sustain<0) _sustain = 0;
        }

        abstract public void apply(float v);

        override public void onTick(float deltaTimeS) {
            if(!getIsTicking())
                return;

            _elapsed += deltaTimeS;
            
            // 0 - off/stopped , 1 - turning on, 2 - on, 3 - turning off
            var fadeProgress = 0f;
           
            if(_phase == 1) {
                fadeProgress = _elapsed / _inTime;
                if(_reached == 0)
                    _current = Mathf.SmoothStep(_default, _target, fadeProgress);
                else
                    _current = Mathf.SmoothStep(_reached, _target, fadeProgress);
            }
            else if(_phase == 3) {
                fadeProgress = _elapsed / _outTime;
                _current = Mathf.SmoothStep(_reached, _default, fadeProgress);
            }

            if(fadeProgress > 1f) {
                // fade in or out finished so...
                if(_sustain > 0)
                    _sustain -= deltaTimeS;
                else {
                    if(_phase == 1) {
                        _reached = _current;
                        _phase = 3; // no (more) sustain, begin fade out
                        _elapsed = 0f;
                    }
                    else {
                        //Log.d($"setting isAlive to false, sustain={_sustain}");
                        _phase = 0; // fadeoutFinished, mark for dispose
                    }
                }
            }

            apply(_current);
        }
    }



    ///////////////////////////////////////////////////
    // SmoothOnOffInterpolator
    // like a fuzzy on/off switch
    ///////////////////////////////////////////////////

    public abstract class OnOffInterpolator : LifecycleInterpolable {
        protected float _inTime; // enable time
        protected float _outTime; // disable time
        protected float _target; // target value
        protected float _offValue;
        protected float _elapsed = 0; // current clock
        protected float _current = 0; // current value
        //protected float _reached = 0; // reached value if in phase was aborted
        protected Func<float, float> _sustain_action = null;

        public OnOffInterpolator(float onValue, float offValue, float inTime, float outTime, Func <float, float> ac = null) {
            _target = onValue;
            _offValue = offValue;
            _inTime = inTime;
            _outTime = outTime;
            _elapsed = 0;
            _sustain_action = ac;
        }

        ~OnOffInterpolator() {

        }

        // 0 - off/stopped , 1 - turning on/paused, 2 - on, 3 - turning off
        public void setState(bool state) {
            if(state && (_phase == 0 || _phase == 3)) {
                _phase = 1;
                _elapsed = 0;
            }
            else if(!state && (_phase == 1 || _phase == 2)) {
                _phase = 3;
                _elapsed = 0;
            }
            //Log.d($"onoff int set to {state}");

        }

        public void reset() {
            apply(_offValue);
        }

        abstract public void apply(float v);

        public void changeMax(float m) {
            _target = m;
        }

        override public void stop() {
            //Log.d("Stop w on off interpolator");
            setState(false);
        }

        override public void onTick(float deltaTimeS) {
            if(!getIsTicking())
                return;

            _elapsed += deltaTimeS;
            
            float fadeProgress = 0;

            // 0 - off, 1 - turning on, 2 - on, 3 - turning off
            if(_phase == 0) {
                _current = _offValue;
            }            
            else if(_phase == 1) {
                fadeProgress = _elapsed / _inTime;
                _current = Mathf.SmoothStep(_offValue, _target, fadeProgress);
                //_reached = _current;
                if(fadeProgress > 1f)
                    _phase = 2;
            }
            else if(_phase == 2) {
                if(_sustain_action != null)
                    _current = _sustain_action(deltaTimeS);
                else
                    _current = _target;
            }
            else if(_phase == 3) {
                fadeProgress = _elapsed / _outTime;
                _current = Mathf.SmoothStep(_target, _offValue, fadeProgress);
                if(fadeProgress > 1f)
                    _phase = 0;
                //Log.d($"Turning off {_current}");
            }

            //Log.d($"Phase:{_phase} Current:{_current}");

            apply(_current);
        }
    }

    ///////////////////////////////////////////////////
    // VectorInterpolator 
    // for things that move
    ///////////////////////////////////////////////////

    public class QuaternionInterpolator : LifecycleInterpolable {
        
        Quaternion _current;
        Quaternion _from;
        Quaternion _to;
        float elapsed = 0;
        float _time;
        Transform transform;
        
        public QuaternionInterpolator(Transform t, Quaternion to, float time) {
            _current = t.rotation;
            _from = t.rotation;
            _to = to;
            _time = time;
            transform = t;
        }

        override public void stop() {
            _phase = 0;
            //_current = _to;
        }

       override public void onTick(float deltaTimeS) {
            if(!getIsTicking())
                return;

            elapsed+=deltaTimeS;
            var frac = elapsed/_time;

            if(frac >= 1) {
                //Log.d("Stopping quat");
                stop();
            }
            else {
                var smoothed = Mathf.SmoothStep(0f, 1f, frac);
                //transform.rotation = Quaternion.Slerp(_from, _to, frac);
                transform.rotation = Quaternion.Slerp(_from, _to, smoothed);
            }
       }
    }

    public class VectorInterpolator : LifecycleInterpolable {
        Vector3 velocity = Vector3.up; // has to have some speed not to be immediately marked dead
        Vector3 _target;
        Transform transform;
        float _time;

        public VectorInterpolator(Transform t, Vector3 target, float speed, float time = 0F) {
            transform = t;
            if(time != 0)
                _time = time;
            else
                _time = Vector3.Distance(t.position, target)/speed;
            _target = target;
            
        }

        override public void stop() {
            _phase = 0;
            velocity = new Vector3();
        }

        override public void onTick(float deltaTimeS) {
            //Log.d("Weksto onTick");
            if(!getIsTicking())
                return;

            transform.position = Vector3.SmoothDamp(transform.position, _target, ref velocity, _time, Mathf.Infinity, deltaTimeS);

            //Log.d($"wektor {transform}");

            if(velocity.magnitude < 0.001)
                stop();
            //Log.d($"on tick, velo={velocity.magnitude}");
            //Log.d($"{velocity.magnitude}");
        }
    }
}