using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;

using System;
using System.Collections;
using System.Collections.Generic;


namespace Amaimon {
    public abstract class Scene {
        protected string voiceCommand = "";

        public Vector3 root = new Vector3();
        public AActor male;
        public AActor female;

        public AAtom lightingRig;
        public Vector3 lightPos;
        public Quaternion lightRot;

        public SceneState state = null;

        //public InterpolablePack interpolables = new InterpolablePack();
        public InterpolablePack physicsInterpolables = new InterpolablePack();
        List<SchedulerJob> scheduler = new List<SchedulerJob>();

        public Scene(AActor m, AActor f) {
            male = m;
            female = f;
        }

        public abstract void onSetupScene();

        public abstract void onExitScene();

        public abstract Scene onRun();

        public float elapsed {
            get {
                return Time.unscaledTime - startTime;
            }
        }

        private float startTime = 0f;

        private float schedulerTimelineCursor = 0;

        public Scene wait(float howLong) {
            schedulerTimelineCursor += howLong;
            return this;
        }

        public Scene waitFromHere(float howLong) {
            schedulerTimelineCursor = elapsed + howLong;
            return this;
        }

        public Scene andThen(Action<Scene> what) {
            var newJob = new SchedulerJob(schedulerTimelineCursor, what);
            scheduler.Add(newJob);
            return this;
        }

        // schedule at absolute time
        public Scene at(float when, Action<Scene> what) {
            var newJob = new SchedulerJob(when, what);
            scheduler.Add(newJob);
            return this;
        }

        // schedule at absolute time from current bookmark
        public Scene after(float howLong, Action<Scene> what) {
            var newJob = new SchedulerJob(schedulerTimelineCursor+howLong, what);
            //Log.d($"At: {schedulerTimelineCursor+howLong-5}");
            scheduler.Add(newJob);
            return this;
        }

        public void registerPhysics(LifecycleInterpolable i) {
            physicsInterpolables.purge();
            physicsInterpolables.list.Add(i);
        }

        // CALLBACK on scene start
        public void setup() {
            lightingRig = Unity.GetAtomById("Lighting_Rigs");

            if(lightPos.magnitude != 0) {
                lightingRig.position = lightPos;
                lightingRig.rotation = lightRot;
            }
            StoryPlugin.ferEachActor(actor => actor.onEnterScene(this));

            Player.singleton.onSentenceSpoken((string spoken) => {
                Log.d($"got sentence {spoken}");
                voiceCommand = spoken;
            });

            onSetupScene();
            startTime = Time.unscaledTime;
        }

        public void onPhysicsTick() {
            physicsInterpolables.onTick(Time.deltaTime);
            StoryPlugin.ferEachActor(actor => actor.onEachPhysicsTick(this));
        }

        public Scene onEachFrame() {
            StoryPlugin.ferEachActor(actor => actor.onEachFrame(this));

            scheduler.ForEach(i => {
                i.launchIfRequired(this);                
            });

            SceneState nextState = (SceneState) state?.changeIf(this);

            if(nextState != null) {
                setState(nextState);
            }
            else {
                state?.onRun(this);
            }
            return onRun();
        }

        public void onExit() {
            try {
                state?.onExit(this);
            } catch (Exception ex) {};
            state = null;

            StoryPlugin.ferEachActor(actor => {
                try {
                    actor?.onExitScene(this);
                } catch (Exception ex) {};
            });

            scheduler.Clear();
            //interpolables?.clear();
            physicsInterpolables?.clear();

            onExitScene();
        }

        public void setState(SceneState newState) {
            if(state == null || newState.GetType() != state.GetType()) {
                Log.d($"Changed state{newState}");
                state?.onExit(this);

                state = newState;

                state.onEnter(this);
                state.setStartTime();
            }
        }
    }   

    public class SimpleScene : Scene {
        public SimpleScene(AActor m, AActor f) : base(m, f) {}
        override public void onSetupScene() {}
        public override Scene onRun() { return null; }
        public override void onExitScene() {}
    }


    class SexSceneBuilder {
        
        protected Vector3 root = new Vector3();
        protected Vector3 lightPos = new Vector3();
        protected Quaternion lightRot = new Quaternion();
        
        protected AActor m;
        protected AActor f;

        public static SexSceneBuilder begin(AActor mm, AActor ff) {
            var sb = new SexSceneBuilder();
            sb.m = mm;
            sb.f = ff;
            return sb;
        } 

        public SexSceneBuilder setRoot(Vector3 r) {
            root = r;
            return this;
        }

        public SexSceneBuilder setLightPos(Vector3 r) {
            lightPos = r;
            return this;
        }

        public SexSceneBuilder setLightRot(float x, float y, float z) {
            lightRot = Quaternion.Euler(x, y, z);
            return this;
        }


	    //public static T FindWithinSamePlugin<T>(MVRScript self) where T : MVRScript
        public SimpleScene build() {
            var newScene = new SimpleScene(m, f);

            // przestało działać, bo rezygnuję z ustawiania przez te pola, trzeba wywoływać funkcję
            newScene.root = root;
            newScene.lightPos = lightPos;
            newScene.lightRot = lightRot;

            return newScene;
        }
    }

}